/*
    Copyright (C) 2008 Kjetil Matheussen
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software 
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

    Note that the license can be changed to a more liberal one on request if
    you have a good reason.
*/


#include <tlsf.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

//#include <gc/private/gcconfig.h> // Need ALIGNMENT
#include "gcconfig.h"

#include <rollendurchmesserzeitsammler.h>
#include <pthread.h>
#include <semaphore.h>

#include "atomicity/atomicity.h"

//#define DEBUG

#define SAFE 1

// Note that setting any of the two below to 1 will most likely decrease the performance significantly
// on multi processor machines. For single processor machines, it shouldn't make much difference I think.
#define PARALLEL_MARK 0
#define PARALLEL_MEMCPY 0



// Haven't checked this one... (MEM_ALIGN is probably 7, is that the correct number to use?)
//#define MEM_ALIGNED(mem) ((((int)mem) & ((int)MEM_ALIGN)) == 0)
// No, it could very well just be 3: (have to ask tlsf author)
#define MEM_ALIGNED(mem) ((((int)mem) & (ALIGNMENT-1)) == 0) // "int" might need to be replaced by something else.

#define ALIGN_MEM(mem) (((int)mem) & (~(ALIGNMENT-1)))
#define ALIGN_MEM128(mem) (((int)mem) & (~127))

#define GC_MAX(a,b) (((a)>(b))?(a):(b))
#define GC_MIN(a,b) (((a)<(b))?(a):(b))

static int num_cpus;

const static int tar_magic_number = 693927941;

static void* atomic_heap;
static size_t atomic_heap_size;

static void* snapshot_heap;
static size_t nonatomic_heap_size;
static size_t snapshot_heap_size;

static tar_mem_t *freelist_nonatomic;
static tar_mem_t *freelist_atomic;

static int num_live=0;


#ifdef DEBUG
#  define D(a) a
static void bug(const char *fmt, ...){
  va_list argp;
  va_start(argp,fmt);
  vprintf(fmt,argp);
  va_end(argp);
}
#else
#  define D(a)
#endif



/**************************************************/
/*************** Various **************************/
/**************************************************/

int tar_find_num_cpus(void){
  int num_cpus=sysconf(_SC_NPROCESSORS_CONF);
  if(num_cpus<1){
    fprintf(stderr,"Unable to determine number of cpus. Assuming 2.\n");
    return 2;
  }
  printf("Note. Rollendurchmesserzeitsammler somehow managed to {right|wrong}fully determine that this machine has %d cpu(s)...\n",num_cpus);
  return num_cpus;
}


static inline void atomic_add(int *mem,int inc){
  __atomic_add(mem,inc);
}

static inline int atomic_add_get_newval(int *mem,int inc){
  int oldval=__exchange_and_add(mem,inc);
  return oldval+inc;
}

static inline int atomic_add_get_oldval(int *mem,int inc){
  return __exchange_and_add(mem,inc);
}

static inline int atomic_inc_round_and_round(int *mem,int size){
  int ret;

  ret=atomic_add_get_oldval(mem,1);


  /*
    The modulo test below works like this:
    if ret == (size*n)-1, where n>=1

    ...and where n could theoretically be >=2 if there has been a long
    break at this point at a previous execution. (better be safe, although
    for size>=(num_cpus*10), I have my doubts...)
  */
  if( 0 == ( (1+ret) % size))
    atomic_add(mem,-size);

  while(ret>=size)
    ret-=size;

  return ret;
}



/**************************************************/
/*************** Ringbuffer ***********************/
/**************************************************/

/* This ringbuffer handles the situation where there
   are more than one reader and one or more writers.
   Plus that there is no waiting either in the reader
   or writer. Ie. it may lock, but it will never wait
   using for instance the usleep() function.
*/

#define inc_semaphore sem_post
#define dec_semaphore sem_wait

struct ringbuffer{
  char *data;
  int size;
  int datasize;
  int read;
  int write;
  sem_t data_ready;
  //int num_free_elements;
  sem_t unread;
} typedef ringbuffer_t;


static inline void rb_read(ringbuffer_t *rb,void *dest){
  int read;
  sem_wait(&rb->data_ready);

  // Atomically get and set the rb->read variable. (no locks)
  read=atomic_inc_round_and_round(&rb->read,rb->size);
  memcpy(dest,&rb->data[read*rb->datasize],rb->datasize);

  //atomic_add(&rb->num_free_elements,1);
  sem_post(&rb->unread);
}

static inline bool rb_write(ringbuffer_t *rb,void *src,bool multiple_writers,bool can_block){

  if(can_block)
    sem_wait(&rb->unread); // Ouch, can't do this when marking. (might deadlock) (why not just use trywait instead of atomic stuff?)
  else
    if(sem_trywait(&rb->unread)==-1)
      return false;

  //                     // Yes, that will work fine. sem_trywait seems to be just an atomic decrement operation. http://www.koders.com/c/fid65FA3D2FC99DA8F14430BA6CEA5BAA322F8E9F58.aspx

#if 0
  {
    int num_free_elements=atomic_add_get_oldval(&rb->num_free_elements,-1);
    
    if(num_free_elements <= 0){
      atomic_add(&rb->num_free_elements,1);
      return false;
    }
  }
#endif

  {
    int write;  

    if(multiple_writers){
      write=atomic_inc_round_and_round(&rb->write,rb->size);
    }else{
      write=rb->write;
      rb->write++;
      if(rb->write==rb->size)
	rb->write=0;
    }
    
    //fprintf(stderr,"Writing %d\n",write);
    memcpy(&rb->data[write*rb->datasize],src,rb->datasize);
  }

  sem_post(&rb->data_ready);

  return true;
}

#if 0
static bool rb_all_jobs_done(ringbuffer_t *rb){
  return rb->num_free_elements==rb->size-2 ? true : false;
  //  return rb->write==rb->read?true:false;
}
#endif

static ringbuffer_t *rb_new(int size,int datasize){
  ringbuffer_t *rb=calloc(1,sizeof(ringbuffer_t));
  rb->data=calloc(datasize,size);
  rb->size=size;
  rb->datasize=datasize;

  sem_init(&rb->data_ready,0,0);
  sem_init(&rb->unread,0,size-2);
  //rb->num_free_elements = size-2;

  return rb;
}



/**************************************************/
/********** Parallell memcpy **********************/
/**************************************************/

/*
Currently, its performance is horrible. I think its because of 
"false sharing". See http://cobweb.ecn.purdue.edu/~pplinux/PPHOWTO/pphowto-2.html
(that's why PARALLEL_MEMCPY is 0 by default)
*/


static int pm_num_unfinished_jobs=0;
static ringbuffer_t *pm_rb;

static sem_t pm_job_finished;

struct pm_job{
  char *dest;
  char *src;
  size_t size;
};

#if 0
static void my_memcpy(char *dest,char *src,size_t size){
  int i;
  for(i=0;i<size;i++){
    dest[i]=src[i];
  }
}

static void my_memcpy2(char *dest,char *src,size_t size){
  memcpy(dest,src,size);
}
#endif

#if 0
// From intel documentation.
static void *memcpy3(void *restrict b, const void *restrict a, size_t n){
    char *s1 = b;
    const char *s2 = a;
    for(; 0<n; --n)*s1++ = *s2++;
    return b;
}
#endif


#if PARALLEL_MEMCPY
static void pm_dothecopy(struct  pm_job *job){
  //my_memcpy(job->dest,job->src,job->size);
  //my_memcpy2(job->dest,job->src,job->size);
  //memcpy(job->dest,job->src,job->size);
  //newmemcpy(job->dest,job->src,job->size);
  //memcpy3(job->dest,job->src,job->size);
  memcpy(job->dest,job->src,job->size);
  //opt_copy(job->dest,job->src,job->size/sizeof(float));

  if( 1 == atomic_add_get_oldval(&pm_num_unfinished_jobs,-1)){
    sem_post(&pm_job_finished);
  }
}

static void pm_worker(void){
  struct pm_job job;
  static int worker_num=0;
  int this_worker_num=worker_num++;
  printf("pm wrker startd %d\n",this_worker_num);
  while(1){
    rb_read(pm_rb,&job);
    //printf("pm_worker %d: %p %p %d (next: %p)\n",this_worker_num,job.dest,job.src,job.size,job.dest+job.size);
    pm_dothecopy(&job);
  }
}
#endif



static void parallel_memcpy(void *vdest,void **vsrcs, size_t *lens,int num_srcs){
  int blocksize;
  char *dest=vdest;
  int i=0;

  if(!PARALLEL_MEMCPY || num_cpus==1){
    char *dest=vdest;
    for(i=0;i<num_srcs;i++){
      memcpy(dest,vsrcs[i],lens[i]);
      dest+=lens[i];
    }
    return;
  }
  
  // First set blocksize.
  size_t num_bytes=0;
  {

    for(i=0;i<num_srcs;i++)
      num_bytes+=lens[i];

    blocksize=ALIGN_MEM128(GC_MAX(1024*32,num_bytes/(num_cpus*3)));
  }


  // Then find number of jobs.
  {
    pm_num_unfinished_jobs = 0;
    
    for(i=0;i<num_srcs;i++){
      size_t len=lens[i];
      pm_num_unfinished_jobs += (len/blocksize);
      if(len % blocksize)
	pm_num_unfinished_jobs++;
    }
  }
 

  // And finally schedule the jobs

  for(i=0;i<num_srcs;i++){

    char *src=vsrcs[i];
    size_t len=lens[i];
    char *end=src+len;

    while(src<end){
      char *next_src=GC_MIN(end,src + blocksize);
      size_t len=next_src - src;

      //printf("blocksize: %d %d\n",num_bytes,blocksize);

      struct pm_job job;
      
      {
	job.src=src;
	job.dest=dest;
	job.size=len;
      }

      rb_write(pm_rb,&job,false,true);

      dest+=len;
      src=next_src;
    }
  }

  sem_wait(&pm_job_finished);
}



static void init_parallel_memcpy(tar_threadcreator realtime_creator){
  sem_init(&pm_job_finished,0,0);
  pm_rb=rb_new(num_cpus*16,sizeof(struct pm_job));
#if PARALLEL_MEMCPY
  if(num_cpus>1){
    int lokke;
    for(lokke=0;lokke<num_cpus;lokke++){
      realtime_creator(pm_worker);
    }
  }
#endif
}







/**************************************************/
/********** Taking snapshot ***********************/
/**************************************************/

// This function takes a snapshot of the root sets and the dynamic memory area.
//   It returns the size of the copied root sets.
static int take_snapshot(tar_heap_t *heap){

  int num_src=heap->phms_place;
  void *srcs[num_src+1];
  size_t sizes[num_src+1];

  size_t roots_size=0;

  srcs[0]=heap->heap;
  sizes[0]=nonatomic_heap_size;

  // First check size and see if the phm area is too small and needs to be reallocated.
  //   In case it is too small, reallocate and then return NULL.
  {
    size_t duplicate_size;
    {
      int i;
      for(i=0;i<num_src;i++){
	size_t size;
	char *start,*end;
	start=heap->phms_start[i];
	end=heap->phms_end[i];
	size=end-start;
	srcs[i+1]=start;
	sizes[i+1]=size;
	roots_size+=size;
      }
    }

    duplicate_size = roots_size + nonatomic_heap_size;

    // snapshot heap too small. Reallocing.
    if(duplicate_size > snapshot_heap_size){
#if 1
      fprintf(stderr,"Garbage collector out of root space. Please allocate more memory for the roots. (%d/%d/%d)\n",
	      duplicate_size,roots_size,nonatomic_heap_size);
      abort();
#else      
      D(printf("reallocing snapshot %d %d\n",duplicate_size,snapshot_heap_size));
      set_gc_thread_to_non_realtime();
      snapshot_heap_size = nonatomic_heap_size + (roots_size*2);
      snapshot_heap = realloc(snapshot_heap,snapshot_heap_size);
      if(snapshot_heap==NULL){
	fprintf(stderr,"malloc returned NULL. exiting. size: %d. num_live: %d\n",snapshot_heap_size,num_live);
	exit(0);
      }
      return NULL;
#endif
    }
  }

  // Then copy
  {

    parallel_memcpy(snapshot_heap,srcs,sizes,num_src+1);

    return roots_size;
  }
}


void tar_add_root(tar_heap_t *heap,void *vstart,void *vend){
  char *start=vstart;
  char *end=vend;
  int place=heap->phms_place;

  if(end < start){
    char *temp=end;
    end=start;
    start=temp;
  }

  heap->phms_start[place]=start;
  heap->phms_end[place]=end;
  heap->phms_place=place+1;
}





/**************************************************/
/****************** Mark  *************************/
/**************************************************/

static tar_heap_t *mark_heap;


// This can be slow. Should ask tlsf author(s) if theres a better way to
// check if *mem is a valid allocated memory pointer.
// If not, use hash table.
static bool mark_is_mem_valid(tar_mem_t *mem,bool is_atomic){
  tar_mem_t *list;

  if(is_atomic==true)
    list=mark_heap->atomic_mem_gc;
  else
    list=mark_heap->mem_gc;

  while(1){
    if(list==mem)
      return true;
    if(list==NULL)
      return false;
    list=list->next;
  }
}


static void mark_area(char *start,char *end);
#if PARALLEL_MARK
static void push_mark_job(char *start,char *end);
#endif

static void mark(char *raw_mem,bool is_atomic){
  tar_mem_t *mem=(tar_mem_t*)(raw_mem-sizeof(tar_mem_t));
  if(1
     && mem->marked==0
     && mem->magic_number == tar_magic_number
     && mark_is_mem_valid(mem,is_atomic) // Just in case.
     )
    {
      if(is_atomic==true){
	mem->marked=1;
      }else{
	if(atomic_add_get_oldval(&mem->marked,1)==0){ 
	  char *start = raw_mem + ( ((char*)snapshot_heap) - ((char*)mark_heap->heap));
	  char *end   = start + mem->size;
#if PARALLEL_MARK
	  // Currently, my feeling is that this will most likely be slower, even on SMP machines.
	  push_mark_job(start,end);
#else
	  mark_area(start,end);
#endif
	}
      }
    }
}



static void mark_area(char *start,char *end){
  char *heapstart = ((char*)mark_heap->heap)    + sizeof(tar_mem_t);
  char *heapend   = ((char*)mark_heap->heap)    + nonatomic_heap_size;
  char *atomic_heapstart = ((char*)atomic_heap) + sizeof(tar_mem_t);
  char *atomic_heapend   = ((char*)atomic_heap) + atomic_heap_size;
  char *i;

  for(i=start;i<end;i+=ALIGNMENT){  
    //for(i=start-4;i<end;i+=1){
    char *raw_mem=(char*)*(int*)i;

    if(MEM_ALIGNED(raw_mem)){
      if(1
	 && raw_mem >= heapstart
	 && raw_mem <  heapend
	 ){
	mark(raw_mem,false);
      }else{
	if(1
	   && raw_mem >= atomic_heapstart
	   && raw_mem <  atomic_heapend
	   ){
	  mark(raw_mem,true);
	}
      }
    }
  }
}


struct mark_job{
  char *start;
  char *end;
};

static ringbuffer_t *mark_rb;
static sem_t mark_finished;

static int num_mark_jobs_running=0;

static void mark_thread(void){
  struct mark_job job;
  while(1){
    rb_read(mark_rb,&job);

    D(bug("%d: Mark thread got something %p %p %d\n",num_mark_jobs_running,job.start,job.end,job.end-job.start));

    {
      mark_area(job.start,job.end);
    }

    if(1 == atomic_add_get_oldval(&num_mark_jobs_running,-1)){
      D(bug("Mark job finished\n"));
      sem_post(&mark_finished);
    }
  }
}

#if PARALLEL_MARK
static void push_mark_job(char *start,char *end){
  {
    struct mark_job job={start,end};
    
    atomic_add(&num_mark_jobs_running,1);    
    if(rb_write(mark_rb,&job,true,false)==false){
      atomic_add(&num_mark_jobs_running,-1);
      mark_area(start,end);
    }
  }
}
#endif

static void start_mark_job(tar_heap_t *heap,char *start,char *end){
  mark_heap=heap;
  {
    struct mark_job job={start,end};
    atomic_add(&num_mark_jobs_running,1);
    rb_write(mark_rb,&job,false,true);
  }
  sem_wait(&mark_finished); // A forced priority inversion.
  D(bug("  Finished marking.\n"));
}

static void init_mark(tar_threadcreator mark_creator){
  
  sem_init(&mark_finished,0,0);
  
  mark_rb=rb_new(num_cpus*8,sizeof(struct mark_job));
#if PARALLEL_MARK
  {
    int lokke;
    for(lokke=0;lokke<num_cpus;lokke++){
      mark_creator(mark_thread);
    }
  }
#else
  mark_creator(mark_thread);
#endif
}





/**************************************************/
/***************Sweep and free ********************/
/**************************************************/


// Returns true in case there is more to do.
static bool free_a_little(tar_mem_t **freelist,void *tlsf_heap){
  if(*freelist!=NULL){
    tar_mem_t *next=(*freelist)->next;
    free_ex(*freelist,tlsf_heap);
    num_live--;
    D(bug("I just freed (for real) %p\n",*freelist));
    *freelist=next;
    return
      next==NULL
      ? false
      : true;
  }else{
    return false;
  }
}


// I'm pretty sure this one is correct now.
//   Hmm, its probably better to free the memory here as well.
// Returns a pointer to the last non-freed element, or NULL if the
//   list of non-freed elements is empty after sweeping.
static tar_mem_t *sweep(tar_mem_t **freelist,tar_mem_t **root){
  tar_mem_t *prev=NULL;
  tar_mem_t *mem=*root;
  while(mem!=NULL){
    tar_mem_t *next = mem->next;

    if(mem->marked==0){
      mem->next=*freelist;
      *freelist=mem;
      if(prev==NULL){
	*root=next;
      }else{
	prev->next=next;
      }
    }else{
      mem->marked=0;
      prev=mem;
    }
    mem=next;
  }
  return prev;
}


static volatile bool new_gc_can_run=true;

static volatile bool please_pause_freeing=false;
static volatile bool free_is_pausing=false;
static pthread_mutex_t free_mutex = PTHREAD_MUTEX_INITIALIZER;
static sem_t freeing_ready_to_go;

static void sweep_and_free(tar_heap_t *heap){

  heap->mem_gc_last        = sweep(&freelist_nonatomic, &heap->mem_gc);     
  heap->atomic_mem_gc_last = sweep(&freelist_atomic,    &heap->atomic_mem_gc);

  pthread_mutex_lock(&free_mutex);
  {
    while(
	     free_a_little(&freelist_nonatomic,heap->heap)==true
	  || free_a_little(&freelist_atomic,atomic_heap)==true
	  ){
      if(please_pause_freeing==true){
	free_is_pausing=true;
	pthread_mutex_unlock(&free_mutex);
	sem_wait(&freeing_ready_to_go);
	pthread_mutex_lock(&free_mutex);
      }
    }
  }
  free_is_pausing=false;
  pthread_mutex_unlock(&free_mutex);

}


// Tells free to pause in case it is freeing.
static void pause_freeing(void){
  please_pause_freeing=true;
  pthread_mutex_lock(&free_mutex); // Wait for it.
}

// Returns false in case it wasn't necessary to continue freeing.
static bool continue_freeing(void){
  please_pause_freeing=false;
  pthread_mutex_unlock(&free_mutex);
  if(free_is_pausing){
    sem_post(&freeing_ready_to_go);
    return true;
  }
  return false;
}

static void init_freeing(void){
  sem_init(&freeing_ready_to_go,0,0);
}




/**************************************************/
/****************** GC Thread *********************/
/**************************************************/


// Not really necessary to do this before the snapshot, but its put here, because
// earlier only the used part of the heap was duplicated, and maybe later there
// will be an option for doing that again.
//  Hmm, this can be done in sweep/free instead, if sweep and free is merged into one.
static void before_snapshot(tar_heap_t *heap){

  if(heap->mem_gc_last != NULL){ // Also means that mem_gc is not NULL
    heap->mem_gc_last->next = heap->mem;
    heap->mem               = heap->mem_gc;
  }

  if(heap->atomic_mem_gc_last != NULL){
    heap->atomic_mem_gc_last->next = heap->atomic_mem;
    heap->atomic_mem               = heap->atomic_mem_gc;
  }

}


static void after_snapshot(tar_heap_t *heap){

  heap->mem_gc=heap->mem;
  heap->mem=NULL;
  
  heap->atomic_mem_gc=heap->atomic_mem;
  heap->atomic_mem=NULL;
  
  heap->num_allocs=0;
}



static sem_t gc_ready;
static sem_t gc_running;
static volatile bool audio_thread_can_run=true;

static tar_heap_t *main_heap;

static void gc_thread_func(void){
  int roots_size;
  tar_heap_t *heap;

  while(1){

    //D(bug("Gc is waiting.\n"));

    new_gc_can_run=true;
    {
      int ret=sem_wait(&gc_ready);
      if(ret!=0){
        fprintf(stderr,"rollendurchmesserzeitsammler error: sem_wait failed: %d (EINTR=%d, EINVAL=%d)\n",ret,EINTR,EINVAL);
        continue;
      }
    }

    {
      heap=main_heap;
      new_gc_can_run=false;
      audio_thread_can_run=false;
    }
    sem_post(&gc_running);

    {
      before_snapshot(heap);
      roots_size=take_snapshot(heap);
      after_snapshot(heap);
    }
    audio_thread_can_run=true;

    {
      char *roots_start=((char*)snapshot_heap)+nonatomic_heap_size;
      char *roots_end=roots_start+roots_size;
      D(bug("Starting to mark. Start: %p, end: %p, size: %d\n",roots_start,roots_end,roots_size));
      if(1 || heap->num_allocs > 0)
	start_mark_job(heap,roots_start,roots_end);
      sweep_and_free(heap);
    }
  }
}


// Returns false in case it failed.
bool tar_entering_audio_thread(tar_heap_t *heap){

  if(audio_thread_can_run==false)
    return false;

  pause_freeing();
  return true;
}


// Returns true in case the garbage collector is ready to collect
// new garbage.
bool tar_leave_audio_thread(tar_heap_t *heap){
  // continue_freeing() must be called first, even if new_gc_can_run is true.
  if(continue_freeing()==true){
    return false;
  }

  if(new_gc_can_run==false)
    return false;

#if 1
  // Not sure about this one.
  // Well, at least mark and sweep shouldn't have to run if num_allocs is 0. Thats really wasted cycles.
  if(heap->num_allocs == 0)
    return false;
#endif

  heap->phms_place=0;
  return true;
}



void tar_run_gc(tar_heap_t *heap){
  main_heap=heap;
  //fprintf(stderr,"Calling sem_post for gc_ready\\n");
  sem_post(&gc_ready);
  sem_wait(&gc_running);
}

static void init_gc_thread(tar_threadcreator realtime_creator){
  sem_init(&gc_ready,0,0);
  sem_init(&gc_running,0,0);

  realtime_creator(gc_thread_func);
}


/**************************************************/
/****************** Init/create/delete ************/
/**************************************************/


void tar_init(size_t das_atomic_heap_size,
	       size_t das_nonatomic_heap_size,
	       size_t roots_size,
	       //int das_num_cpus,
	       tar_threadcreator realtime_creator,
	       tar_threadcreator mark_creator
	       ){
  nonatomic_heap_size=das_nonatomic_heap_size;
  atomic_heap_size = das_atomic_heap_size;
  num_cpus=tar_find_num_cpus();//das_num_cpus;

  init_parallel_memcpy(realtime_creator);

  atomic_heap=malloc(atomic_heap_size);
  init_memory_pool(atomic_heap_size,atomic_heap);

  snapshot_heap_size=nonatomic_heap_size + roots_size;
  snapshot_heap=malloc(snapshot_heap_size);

  init_mark(mark_creator);
  init_freeing();
  init_gc_thread(realtime_creator);
}



tar_heap_t *tar_new_heap(void){
  size_t size=nonatomic_heap_size;

  tar_heap_t *heap=calloc(1,sizeof(tar_heap_t));

  heap->heap=malloc(size); // Hmm, in case calloc is implemented by using copy-on-write.
  memset(heap->heap,0,size);

  init_memory_pool(size,heap->heap); // BTW. init_memory_pool() returns free mem available, not the physical heap size.

  return heap;
}

void tar_delete(tar_heap_t *heap){

  D(bug("freeing a little\n"));
  //while(free_a_little(heap)==true);
  D(bug("num_live before exit: %d (not necessarily supposed to be 0, but it shouldn't bee too high either)\n",num_live));

  free(heap->heap);
  free(heap);
}




/**************************************************/
/****************** Allocation ********************/
/**************************************************/



// Something may need to be done here. The size of tar_mem needs to be aligned in 8 or 16 (I think),

static void *alloc_common(tar_heap_t *heap,tar_mem_t *mem,size_t size){
  mem->size=size;
  num_live++;
  mem->magic_number=tar_magic_number;
  heap->num_allocs++;  
  return mem+1; // Hmm, yes. Well.
}

void *tar_alloc(tar_heap_t *heap,size_t size){
  //D(bug("hepp2\n"));
  tar_mem_t *mem=calloc_ex(1,size+sizeof(tar_mem_t),heap->heap); // Using calloc_ex to null out memory. (don't want false pointers in the mem)
  //tar_mem_t *mem=calloc(1,size+sizeof(tar_mem_t));
  if(mem==NULL){
    fprintf(stderr,"Out of memory\n");
    return NULL;
  }
  mem->next=heap->mem;
  heap->mem=mem;
  D(bug("alloced something (dyn): %p (%p) (num_allocs: %d / %d) (heap: %p)\n",mem,mem+1,heap->num_allocs,num_live,heap->heap));
  return alloc_common(heap,mem,size);
}

void *tar_alloc_atomic(tar_heap_t *heap,size_t size){
  tar_mem_t *mem=malloc_ex(size+sizeof(tar_mem_t),atomic_heap); // atomic mem are not checked for pointers, so it doesn't have to be nulled out.
  //tar_mem_t *mem=malloc(size+sizeof(tar_mem_t));
  if(mem==NULL){
    fprintf(stderr,"Out of memory\n");
    return NULL;
  }
  mem->next=heap->atomic_mem;
  heap->atomic_mem=mem;
  mem->marked=0;
  D(bug("alloced something (atom): %p (%p) (num_allocs: %d / %d) (heap: %p)\n",mem,mem+1,heap->num_allocs,num_live,heap->heap));
  return alloc_common(heap,mem,size);
}



void *tar_malloc_atomic_uncollectable(size_t size){ // familiar name from hbgc
  return malloc_ex(size,atomic_heap);
}

void *tar_calloc_atomic_uncollectable(size_t size1,size_t size2){ // familiar name from hbgc
  return calloc_ex(size1,size2,atomic_heap);
}

void tar_free_atomic_uncollectable(void *mem){
  free_ex(mem,atomic_heap);
}






#if TEST
void* threadstart_realtime(void *arg){
  tar_threadstart threadstart=(tar_threadstart)arg;
  threadstart();
  return NULL;
}

void create_realtime_thread(tar_threadstart threadstart){
  pthread_t *thread;
  thread=calloc(sizeof(pthread_t*),1);
  pthread_create(thread,NULL,threadstart_realtime,threadstart);
}

void* threadstart_mark(void *arg){
  tar_threadstart threadstart=(tar_threadstart)arg;
  threadstart();
  return NULL;
}

void create_mark_thread(tar_threadstart threadstart){
  pthread_t *thread;
  thread=calloc(sizeof(pthread_t*),1);
  pthread_create(thread,NULL,threadstart_mark,threadstart);
}


int main(){
  //int stack_start;
  tar_heap_t *tar;
  tar_init(1024*1024,
	   1024*1024,
	   1024*1024,
	   create_realtime_thread,
	   create_mark_thread
	   );
  {
    tar=tar_new_heap();
  }

#if 0
  int val=200;
  int val2=__exchange_and_add(&val,-1);
  int val3=atomic_add_get_newval(&val,100);
  // result now: val=299, val2=200, val3=299
  printf("memsize: %d, size: %d %d (%d %d %d)\n",sizeof(tar_mem_t),sizeof(pthread_mutex_t),sizeof(size_t),val,val2,val3); // 24 is too much to put into tar_mem_t.
  char *mem=tar_alloc_atomic(tar,100);
  printf("Alloc something: %p %p %p %p %p\n",mem,tar->atomic_mem,tar->atomic_mem->next,tar->atomic_mem_gc,freelist_nonstatic);
  exit(0);
#endif

#if 0
  void *mem=calloc_ex(2,50,tar->heap);
  D(bug("heapstart: %p, mem: %p\n",tar->heap,mem));
  return 0;
#endif

  //int num_frames=1024;
  int num_extra=0;

  while(1){
    if(tar_entering_audio_thread(tar)==false){
      //D(bug("Using too much cpu. Skipping.\n"));
      usleep(50);
      continue;
    }
    D(bug("hupp6\n"));
    int frame;
    void **mem;
    for(frame=0;frame<5000*((float)random()/(float)RAND_MAX);frame++){
      mem=tar_alloc(tar,sizeof(void*)*128);
      if(mem==NULL){
	printf("GAKK mem=NULL, num_live: %d %d\n",num_live,num_extra);
	exit(0);
      }
      mem[0]=tar_alloc_atomic(tar,100);
      if(mem[0]==NULL){
	printf("GAKK mem[0]=NULL, num_live: %d %d\n",num_live,num_extra);
	exit(0);
      }		   
      mem[20]=tar_alloc(tar,20);
      num_extra+=sizeof(void*)*128;
      usleep(1);
      /* ... */
    }
    D(bug("GAKK **mems address: %p, mems val: %p\n",&mem,mem));
    if(tar_leave_audio_thread(tar)){
      tar_add_root(tar,&mem,(&mem)+50);
      //tar_add_set(tar,&globals->pointers_start,&globals->pointers_end);
      D(bug("hupp4\n"));
      tar_run_gc(tar);
    }
    usleep(100000);
    //sleep(2);
    if(frame==3)
      break;
  }
  tar_delete(tar);
  return 0;
}
#endif

